






createUniqueColumnName <- function(x)
{
    make.unique(c(names(x), "sample"))[ncol(x) + 1]
}


## FIXME: should have some way to combine or transform channel, e.g.
## densityplot(factor(Visit) ~ log(`FSC-H`) | factor(Patient), GvHD)
## Unfortunately, that won't be straightforward (nor that important probably)



setMethod("densityplot",
          signature(x = "formula", data = "flowSet"),
          function(x, data, xlab,
                   as.table = TRUE, overlap = 0.3, 
                   ...)
      {
          pd <- pData(phenoData(data))
          uniq.name <- createUniqueColumnName(pd)
          ## ugly hack to suppress warnings about coercion introducing
          ## NAs (needs to be `undone' inside prepanel and panel
          ## functions):
          pd[[uniq.name]] <- factor(sampleNames(data)) 
          channel <- x[[3]]  
          if (length(channel) == 3)
          {
              channel <- channel[[2]]
              x[[3]][[2]] <- as.name(uniq.name)
          }
          else x[[3]] <- as.name(uniq.name)
          channel <- as.character(channel)

          prepanel.densityplot.flowset <- 
              function(x, y, darg = list(n = 30),
                       frames, channel,
                       overlap = 0.3,
                       ...)
              {
                  xl <- numeric(0)
                  for (nm in as.character(x))
                  {
                      xx <- exprs(frames[[nm]])[, channel]
                      xl <- c(xl, range(xx))
                  }
                  list(xlim = range(xl, finite = TRUE))
              }

          panel.densityplot.flowset <-
              function(x, y, darg = list(n = 30), ref = FALSE,
                       frames, channel,
                       overlap = 0.3,

                       col = superpose.polygon$col,
                       lty = superpose.polygon$lty,
                       lwd = superpose.polygon$lwd,
                       alpha = superpose.polygon$alpha,
                       border = superpose.polygon$border,
                       ...)
              {
                  superpose.line <- trellis.par.get("superpose.line")
                  superpose.polygon <- trellis.par.get("superpose.polygon")
                  reference.line <- trellis.par.get("reference.line")
                  ycode <- as.numeric(y)
                  if (any(duplicated(ycode)))
                  {
                      warning("Some combinations seem to have multiple samples.  \n  Only one will be used.")
                  }
                  ny <- nlevels(y)
                  col <- rep(col, length = ny)
                  lty <- rep(lty, length = ny)
                  lwd <- rep(lwd, length = ny)
                  alpha <- rep(alpha, length = ny)
                  border <- rep(border, length = ny)
                  x <- as.character(x)
                  height <- (1 + overlap)
                  for (i in rev(seq_len(ny)))
                      if (i %in% ycode)
                      {
                          nm <- x[match(i, ycode)]
                          xx <- exprs(frames[[nm]])[, channel]
                          h <- do.call(density, c(list(x = xx), darg))
                          n <- length(h$x)
                          max.d <- max(h$y)
                          panel.polygon(x = h$x[c(1, 1:n, n)],
                                        y = i + height * c(0, h$y, 0) / max.d,
                                        col = col[i], border = border[i],
                                        lty = lty[i], lwd = lwd[i], alpha = alpha[i])
                          if (ref)
                          {
                              panel.abline(h = i,
                                           col = reference.line$col,
                                           lty = reference.line$lty,
                                           lwd = reference.line$lwd,
                                           alpha = reference.line$alpha)
                          }
                      }
              }

          if (missing(xlab)) xlab <- channel
          bwplot(x, data = pd, 

                 prepanel = prepanel.densityplot.flowset,
                 panel = panel.densityplot.flowset,

                 frames = data@frames,
                 channel = channel,
                 as.table = as.table,
                 overlap = overlap,

                 xlab = xlab,

                 lattice.options = list(axis.padding = list(factor = c(0.6, 1 + 2 * overlap))),
                 horizontal = TRUE, ...)
      })








setMethod("xyplot",
          signature(x = "formula", data = "flowSet"),
          function(x, data, xlab, ylab,
                   as.table = TRUE,
                   pch = ".", smooth = TRUE,
                   ...)
      {
          pd <- pData(phenoData(data))
          uniq.name <- createUniqueColumnName(pd)
          ## ugly hack to suppress warnings about coercion introducing
          ## NAs (needs to be `undone' inside prepanel and panel
          ## functions):
          pd[[uniq.name]] <- factor(sampleNames(data)) 
          channel.y <- x[[2]]
          channel.x <- x[[3]]
          if (length(channel.x) == 3)
          {
              channel.x <- channel.x[[2]]
              x[[3]][[2]] <- as.name(uniq.name)
              x[[2]] <- NULL
          }
          else
          {
              x[[3]] <- as.name(uniq.name)
              x[[2]] <- NULL
          }
          channel.x <- as.character(channel.x)
          channel.y <- as.character(channel.y)

          prepanel.xyplot.flowset <- 
              function(x, 
                       frames, channel.x, channel.y,
                       ...)
              {
                  if (length(nm <- as.character(x)) > 1)
                      stop("must have only one flow frame per panel")
                  if (length(nm) == 1)
                  {
                      xx <- exprs(frames[[nm]])[, channel.x]
                      yy <- exprs(frames[[nm]])[, channel.y]
                      list(xlim = range(xx, finite = TRUE),
                           ylim = range(yy, finite = TRUE),
                           dx = diff(xx), dy = diff(yy))
                  }
                  else list()
              }

          panel.xyplot.flowset <-
              function(x, 
                       frames, channel.x, channel.y,

                       pch, smooth,
                       ...)
              {
                  x <- as.character(x)
                  if (length(x) > 1) stop("must have only one flow frame per panel")
                  
                  for (nm in x)
                  {
                      xx <- exprs(frames[[nm]])[, channel.x]
                      yy <- exprs(frames[[nm]])[, channel.y]
                      
                      if (smooth) panel.smoothScatter(xx, yy, ...)
                      else panel.xyplot(xx, yy, pch = pch, ...)
                  }
              }

          if (missing(xlab)) xlab <- channel.x
          if (missing(ylab)) ylab <- channel.y
          densityplot(x, data = pd, 

                      prepanel = prepanel.xyplot.flowset,
                      panel = panel.xyplot.flowset,

                      frames = data@frames,
                      channel.x = channel.x,
                      channel.y = channel.y,
                      as.table = as.table,

                      xlab = xlab,
                      ylab = ylab,
                      pch = pch, smooth = smooth,

                      ...)
      })




setMethod("levelplot",
          signature(x = "formula", data = "flowSet"),
          function(x, data, xlab, ylab,
                   as.table = TRUE, contour = TRUE, labels = FALSE,
                   n = 50, 
                   ...)
      {
          samples <- sampleNames(data)
          channel.y <- x[[2]]
          channel.x <- x[[3]]
          if (length(channel.x) == 3) ## cy ~ cx | foo+bar
          {
              channel.x <- channel.x[[2]]
              my.formula <- x
              my.formula[[2]] <- as.name("z")
              my.formula[[3]][[2]] <- (z ~ x * y)[[3]]
          }
          else ## cy ~ cx
          {
              my.formula <- z ~ x * y
          }
          channel.x <- as.character(channel.x)
          channel.y <- as.character(channel.y)
          range.x <- range(unlist(lapply(samples, function(nm) range(exprs(data@frames[[nm]])[, channel.x], finite = TRUE))))
          range.y <- range(unlist(lapply(samples, function(nm) range(exprs(data@frames[[nm]])[, channel.y], finite = TRUE))))
          ## range.y <- range(unlist(eapply(data@frames, function(x) range(x[, channel.y], finite = TRUE))))
          range.x <- lattice:::extend.limits(range.x, prop = 0.05)
          range.y <- lattice:::extend.limits(range.y, prop = 0.05)
          computeKde <- function(nm)
          {
              xx <- exprs(data@frames[[nm]])[, channel.x]
              yy <- exprs(data@frames[[nm]])[, channel.y]
              temp <- kde2d(xx, yy, n = n, lims = c(range.x, range.y))
              data.frame(x = rep(temp$x, n),
                         y = rep(temp$y, each = n),
                         z = as.vector(temp$z))
          }
          ft <- do.call(make.groups, sapply(samples, computeKde, simplify = FALSE))
          ## The next step is not efficient, as it replicates the
          ## phenodata many times over.  However, bypassing this is a
          ## task for another day
          ft <- cbind(ft, pData(phenoData(data))[as.character(ft$which), , drop = FALSE])
          if (missing(xlab)) xlab <- channel.x
          if (missing(ylab)) ylab <- channel.y
          levelplot(my.formula, data = ft,
                    contour = contour, labels = labels,
                    xlab = xlab,
                    ylab = ylab,
                    as.table = as.table, 
                    ...)
      })


##Example:
##
## require(colorspaces)
## YlOrBr <- c("#FFFFD4", "#FED98E", "#FE9929", "#D95F0E", "#993404")
## colori <- colorRampPalette(YlOrBr, space = "Lab")
## levelplot(`SSC-H` ~ `FSC-H` | name, samp, col.regions=colori(50), main="Contour Plot")
##




## QQ plot

## setGeneric("qqmath")



setMethod("qqmath",
          signature(x = "formula", data = "flowSet"),
          function(x, data, xlab, ylab,
                   f.value = function(n) ppoints(ceiling(sqrt(n))),
                   distribution = qnorm,
                   ...)
      {
          pd <- pData(phenoData(data))
          uniq.name <- createUniqueColumnName(pd)
          ## ugly hack to suppress warnings about coercion introducing NAs
          pd[[uniq.name]] <- factor(sampleNames(data))
          channel <- x[[2]]
          if (length(channel) == 3)
          {
              channel <- channel[[2]]
              x[[2]][[2]] <- as.name(uniq.name)
          }
          else x[[2]] <- as.name(uniq.name)
          channel <- as.character(channel)

          prepanel.qqmath.flowset <- 
              function(x, frames, channel,
                       f.value, distribution,
                       ...)
              {
                  distribution <- if (is.function(distribution)) 
                      distribution
                  else if (is.character(distribution)) 
                      get(distribution)
                  else eval(distribution)

                  nobs <- max(unlist(eapply(frames, function(x) nrow(exprs(x)))))

                  qrange <-
                      if (is.null(f.value)) 
                          ppoints(sqrt(nobs))
                      else if (is.numeric(f.value))
                            f.value
                      else f.value(nobs)
                  qrange <- range(qrange)

                  xx <- distribution(qrange)
                  yy <-
                      unlist(eapply(frames,
                                    function(x) {
                                        quantile(exprs(x)[, channel],
                                                 qrange,
                                                 na.rm = TRUE)
                                    }))

                  dx <- rep(diff(distribution(c(0.25, 0.75))), length(ls(frames)))
                  dy <-
                      unlist(eapply(frames,
                                    function(x) {
                                        diff(quantile(exprs(x)[, channel],
                                                      c(0.25, 0.75),
                                                      na.rm = TRUE))
                                    }))

                  list(xlim = range(xx, finite = TRUE),
                       ylim = range(yy, finite = TRUE),
                       dx = dx, dy = dy)
              }

          panel.qqmath.flowset <-
              function(x, 
                       frames, channel,
                       f.value, distribution,
                       grid = FALSE,

                       col = superpose.symbol$col,
                       col.points = col,
                       pch = superpose.symbol$pch,
                       cex = superpose.symbol$cex,
                       alpha = superpose.symbol$alpha,
                       col.line = col,
                       lty = superpose.line$lty,
                       lwd = superpose.line$lwd,
                       ...)
              {
                  if (grid) panel.grid(h = -1, v = -1)
                  superpose.symbol <- trellis.par.get("superpose.symbol")
                  superpose.line <- trellis.par.get("superpose.line")

                  nx <- length(x)
                  col.points <- rep(col.points, length = nx)
                  col.line <- rep(col.line, length = nx)
                  pch <- rep(pch, length = nx)
                  cex <- rep(cex, length = nx)
                  lty <- rep(lty, length = nx)
                  lwd <- rep(lwd, length = nx)
                  alpha <- rep(alpha, length = nx)

                  distribution <- if (is.function(distribution)) 
                      distribution
                  else if (is.character(distribution)) 
                      get(distribution)
                  else eval(distribution)

                  nobs <- max(unlist(eapply(frames, function(x) nrow(exprs(x)))))
                  qq <-
                      if (is.null(f.value)) 
                          ppoints(sqrt(nobs))
                      else if (is.numeric(f.value))
                          f.value
                      else f.value(nobs)
                  xx <- distribution(qq)
                  x <- as.character(x)
                  for (i in seq_along(x))
                  {
                      nm <- x[i]
                      yy <-
                          quantile(exprs(frames[[nm]])[, channel],
                                   qq, na.rm = TRUE)
                      panel.xyplot(xx, yy,
                                   col.line = col.line[i],
                                   col = col[i],
                                   cex = cex[i],
                                   pch = pch[i],
                                   lty = lty[i],
                                   lwd = lwd[i],
                                   alpha = alpha[i],
                                   ...)
                  }
              }

          if (missing(xlab)) xlab <- deparse(substitute(distribution))
          if (missing(ylab)) ylab <- channel
          qqmath(x, data = pd,
                 f.value = f.value, distribution = distribution,
                 prepanel = prepanel.qqmath.flowset,
                 panel = panel.qqmath.flowset,
                 frames = data@frames,
                 channel = channel,
                 xlab = xlab,
                 ylab = ylab,
                      
                 ...)
      })




setMethod("splom",
          signature(x = "flowFrame", data = "missing"),
          function(x, data, 
                   pscales = 0,
                   smooth = TRUE, pch = ".",
                   time = "Time", exclude.time = TRUE,
                   panel = function(x, y, smooth, ...) {
                       if (smooth) panel.smoothScatter(x, y, ...)
                       else panel.xyplot(x, y, ...)
                   },
                   ...)
      {
          expr <- exprs(x)
          column.names <- colnames(expr)
          if (exclude.time) column.names <- column.names[column.names != time]
          splom(exprs(x[, column.names]),
                pscales = pscales, 
                panel = panel,
                smooth = smooth, pch = pch,
                ...)
      })



setMethod("parallel",
          signature(x = "flowFrame", data = "missing"),
          function(x, data, 
                   reorder.by = function(x) var(x, na.rm = TRUE),
                   time = "Time", exclude.time = TRUE,
                   ...)
      {
          expr <- exprs(x)
          column.names <- colnames(expr)
          if (exclude.time) column.names <- column.names[column.names != time]
          if (!is.null(reorder.by))
          {
              column.order <- rev(order(apply(expr[, column.names], 2, reorder.by)))
              column.names <- column.names[column.order]
          }
          parallel(expr[, column.names],
                   ...)
      })




## xyplot for flowFrames, plotting columns against time

setMethod("xyplot",
          signature(x = "flowFrame", data = "missing"),
          function(x, data, xlab = time, ylab = "", time = "Time",
                   layout,
                   type = "l", smooth = FALSE,
                   ...)
      {
          expr <- exprs(x)
          if (!(time %in% colnames(expr)))
              stop("Name of the Time (X) variable must be specified as the 'time' argument")
          time.x <- expr[, time]
          fakedf <- data.frame(channel = setdiff(colnames(expr), time), time = 1)
          prepanel.xyplot.flowframe <- 
              function(x, y, time.x, expr, ...)
              {
                  xx <- time.x
                  yy <- expr[, as.character(y)]
                  list(xlim = range(xx, finite = TRUE),
                       ylim = range(yy, finite = TRUE),
                       dx = diff(xx), dy = diff(yy))
              }
          panel.xyplot.flowframe <- 
              function(x, y, time.x, expr, smooth = FALSE, ...)
              {
                  xx <- time.x
                  yy <- expr[, as.character(y)]
                  if (smooth) panel.smoothScatter(xx, yy, ...)
                  else panel.xyplot(xx, yy, ...)
              }
          if (missing(layout)) layout <- c(1, ncol(expr) - 1)
          xyplot(channel ~ time | channel, data = fakedf,
                 prepanel = prepanel.xyplot.flowframe,
                 panel = panel.xyplot.flowframe,
                 type = type, smooth = smooth,
                 time.x = time.x, expr = expr,
                 xlab = xlab, ylab = ylab,
                 layout = layout,
                 default.scales = list(y = list(relation = "free", rot = 0)),
                 ...)
      })


## xyplot with a formula.  We'll make this very simple; the drawback
## being that the expression matrix will be copied, the upshot being
## that all the fancy xyplot formula stuff will be valid.

setMethod("xyplot",
          signature(x = "formula", data = "flowFrame"),
          function(x, data,
                   smooth = TRUE, 
                   panel = if (smooth) panel.smoothScatter else panel.xyplot,
                   ...)
      {
          xyplot(x, data = as.data.frame(exprs(data)),
                 smooth = smooth, 
                 panel = panel,
                 ...)
      })







